-file("eqc-1.47.1/examples/counters_eqc.erl", 0).
%%% @author Thomas Arts
%%% @copyright (C) 2019, Quviq AB
%%% @doc Testing the counters module
%%%
%%% @end
%%% Created : 16 May 2019 by Thomas Arts

-module(counters_eqc).

-include_lib("../include/eqc.hrl").
-include_lib("../include/eqc_statem.hrl").

-compile([export_all, nowarn_export_all]).

%% -- State and state functions ----------------------------------------------
initial_state() ->
    #{}.

%% -- Generators -------------------------------------------------------------


%% -- Operations -------------------------------------------------------------

%% --- Operation: new ---
new_args(_S) ->
    [ ?SUCHTHAT(N, nat(), N >= 1), oneof([[], [atomics], [write_concurrency]]) ].

new(Size, Opts) ->
    counters:new(Size, Opts).

new_next(S, Value, [Size, _Opts]) ->
    maps:merge(S, maps:from_list([{{Value, N}, 0} || N <- lists:seq(1, Size)])).



%% --- Operation: add ---
add_pre(S) ->
    S =/= #{}.

add_args(S) ->
    [elements(maps:keys(S)), int()].

add_pre(S, [{Counter, Idx}, Incr]) ->
    %% for shrinking
    maps:is_key({Counter, Idx}, S).

add({Counter, Idx}, Incr) ->
    counters:add(Counter, Idx, Incr).

add_next(S, _Value, [{Counter, Idx}, Incr]) ->
    maps:update_with({Counter, Idx}, fun(X) -> X + Incr end, S).

add_post(_S, _Args, Res) ->
    eq(Res, ok).


%% --- Operation: get ---
read_pre(S) ->
    S =/= #{}.

read_args(S) ->
    ?LET({Counter, Idx}, elements(maps:keys(S)),
         [Counter, Idx]).

read_pre(S, [Counter, Idx]) ->
    maps:is_key({Counter, Idx}, S).

read(Counter, Idx) ->
    counters:get(Counter, Idx).

read_post(S, [Counter, Idx], Res) ->
    eq(Res, maps:get({Counter, Idx}, S)).


weight(S, read) -> length(maps:keys(S));
weight(S, add)  -> length(maps:keys(S));
weight(_, _) -> 1.




%% -- Property ---------------------------------------------------------------
prop_counters() ->
    ?FORALL(Cmds, parallel_commands(?MODULE),
    begin
        {H, S, Res} = run_parallel_commands(Cmds),
        check_command_names(Cmds,
            measure(length, commands_length(Cmds),
                pretty_commands(?MODULE, Cmds, {H, S, Res},
                                Res == ok)))
    end).

